# ******************************************************************************
#
# sigcalc, significant figures calculations
#
# Copyright 2023-2024 Jeremy A Gray <gray@flyquackswim.com>.
#
# All rights reserved.
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# ******************************************************************************

"""Quantity class for significant figure calculations."""

import math
from decimal import Decimal
from decimal import getcontext
from warnings import warn

import mpmath

# Reusable decimal constants.
Zero = Decimal("0")
One = Decimal("1")
Ten = Decimal("10")


def _most_significant_place(num):
    """Find the most significant place of a number.

    Find the most significant place of a number and return it as the
    exponent for a power of ten corresponding to the place value.

    Parameters
    ----------
    num : Decimal
        The number whose most significant place is needed.

    Returns
    -------
    Decimal
        The exponent of the most significant place.
    """
    if num == Zero:
        return Zero

    return Decimal(math.floor(math.log10(abs(num))))


class Quantity:
    """A quantity for significant figure calculations."""

    def __init__(self, value, figures=None, constant=False):
        """Initialize a ``Quantity()``.

        Parameters
        ----------
        value : str
            Value of the quantity, to be converted to a ``Decimal``.
        figures : str
            Significant figures of the quantity, to be converted to a
            ``Decimal``.
        constant : bool
            Set as constant (unlimited precision), or not.
        """
        if not constant and figures is None:
            raise TypeError("`figures` should not be None for non-constants")
        elif constant:
            figures = One

        self.value = Decimal(str(value))

        getcontext().prec = max(
            getcontext().prec, int(figures), len(self.value.as_tuple().digits)
        )

        self.reported = self.value
        if not constant:
            self.figures = Decimal(str(figures))
        else:
            self.figures = getcontext().prec
        self._constant = constant

    # Output operations.
    def __repr__(self):
        """Represent a ``Quantity``."""
        rep = f'Quantity("{self.value}", "{self.figures}"'

        if self.constant:
            rep += f", constant={self.constant}"

        return rep + ")"

    def __format__(self, specifier):
        """Format a ``Quantity`` object.

        Format a ``Quantity`` object, rounding to significant figures
        first and then passing any format specifier through to
        ``Decimal().__format__()`` for formatting.
        """
        return format(self._round(), specifier)

    def __str__(self):
        """Stringify a ``Quantity`` object.

        Stringify a ``Quantity`` object, rounding to significant
        figures first and then using ``Decimal().__str__()`` for
        stringification.
        """
        return str(self._round())

    # Unary operations.
    def __neg__(self):
        """Negate a ``Quantity()`` object."""
        return Quantity(-self.value, self.figures, self.constant)

    def __pos__(self):
        """Return a ``Quantity()`` object."""
        return Quantity(self.value, self.figures, self.constant)

    def __abs__(self):
        """Calculate the magnitude of a ``Quantity()`` object."""
        return Quantity(abs(self.value), self.figures, self.constant)

    # Comparisons.
    def __lt__(self, other):
        """Determine the ordering of two ``Quantity()`` objects."""
        if isinstance(other, Quantity) and self._round() < other._round():
            return True

        return False

    def __le__(self, other):
        """Determine the ordering of two ``Quantity()`` objects."""
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other):
        """Determine equality of two ``Quantity()`` objects."""
        if (
            isinstance(other, Quantity)
            and self._round() == other._round()
            and (self.figures == other.figures or (self.constant and other.constant))
        ):
            return True

        return False

    def __ne__(self, other):
        """Determine the inequality of two ``Quantity()`` objects."""
        return not self.__eq__(other)

    def __gt__(self, other):
        """Determine the ordering of two ``Quantity()`` objects."""
        return not self.__lt__(other) and self.__ne__(other)

    def __ge__(self, other):
        """Determine the ordering of two ``Quantity()`` objects."""
        return not self.__lt__(other) or self.__eq__(other)

    # Arithmetic operations.
    def __add__(self, other):
        """Add two ``Quantity()`` objects."""
        if isinstance(other, Quantity):
            value = self.value + other.value

            if self.constant and other.constant:
                return Quantity(value, getcontext().prec, constant=True)
            elif self.constant and not other.constant:
                least = _most_significant_place(other.value) - other.figures + 1
            elif other.constant and not self.constant:
                least = _most_significant_place(self.value) - self.figures + 1
            else:
                least = max(
                    _most_significant_place(self.value) - self.figures + 1,
                    _most_significant_place(other.value) - other.figures + 1,
                )

            most = _most_significant_place(value)

            return Quantity(value, max(most - least + 1, 1))

        return NotImplemented

    def __sub__(self, other):
        """Subtract two ``Quantity()`` objects."""
        return self + -other

    def __mul__(self, other):
        """Multiply two ``Quantity`` objects."""
        if isinstance(other, Quantity):
            if self.constant and other.constant:
                return Quantity(
                    self.value * other.value, getcontext().prec, constant=True
                )
            elif self.constant and not other.constant:
                return Quantity(self.value * other.value, other.figures)
            elif other.constant and not self.constant:
                return Quantity(self.value * other.value, self.figures)
            else:
                return Quantity(
                    self.value * other.value, min(self.figures, other.figures)
                )

        return NotImplemented

    def __truediv__(self, other):
        """Divide two ``Quantity`` objects."""
        if isinstance(other, Quantity):
            if self.constant and other.constant:
                return Quantity(
                    self.value / other.value, getcontext().prec, constant=True
                )
            elif self.constant and not other.constant:
                return Quantity(self.value / other.value, other.figures)
            elif other.constant and not self.constant:
                return Quantity(self.value / other.value, self.figures)
            else:
                return Quantity(
                    self.value / other.value, min(self.figures, other.figures)
                )

        return NotImplemented

    def __pow__(self, other, modulo=None):
        """Calculate a power of a ``Quantity``.

        Calculate a power of a ``Quantity`` and compute the correct
        number of significant figures.  The ``__pow__`` from
        ``decimal.Decimal`` is used to calculate the value of the
        power.

        Significant figures follows "significance in, significance
        out" rule since this is (naively) multiplication.  This makes
        the assumption that the exponent will usually be an integer
        anyway and that the precision will be controlled by ``self``.

        >>> from sigcalc import Quantity
        >>> a = Quantity("2", "3")
        >>> b = Quantity("2", "3")
        >>> a**b
        Quantity("4", "3")
        >>> b = Quantity("2", "2")
        >>> a**b
        Quantity("4", "3")
        >>> a = Quantity("2", "2")
        >>> b = Quantity("2", "3")
        >>> a**b
        Quantity("4", "2")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed power and significant
            figures and constant value from the input ``Quantity``.
        """
        value = pow(self.value, other.value, modulo)

        if self.constant:
            return Quantity(
                value,
                constant=True,
            )

        return Quantity(
            value,
            self.figures,
        )

    @property
    def constant(self):
        """Mark a ``Quantity`` as a constant with unlimited precision."""
        return self._constant

    @constant.setter
    def constant(self, value):
        """Set a ``Quantity`` object's ``constant`` property.

        Set a ``Quantity`` object's ``constant`` property to value, if
        boolean.
        """
        if isinstance(value, bool):
            self._constant = value

    def _round(self):
        """Round a ``Quantity`` object ``value`` to significant figures.

        Round a ``Quantity`` object ``value`` to significant figures
        using the context rounding mode, returning a ``Decimal``.
        Return ``self.value`` unrounded if ``self.constant`` is
        ``True``.

        Returns
        -------
        Decimal
            The rounded ``value`` of the ``Quantity``.
        """
        if self.constant:
            return self.value

        place = self.value.adjusted() - self.figures + One
        return self.value.quantize(Decimal(f"1e{place}"))

    def round(self):  # dead: disable
        """Round a ``Quantity`` object to significant figures.

        Round a ``Quantity`` object to significant figures using the
        context rounding mode, returning a new ``Quantity`` with only
        the rounded significant figures.  Return a new ``Quantity``
        equal to ``self`` unrounded if ``self.constant`` is ``True``.

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` equal to the rounded ``self``.
        """
        if self.constant:
            return Quantity(self.value, self.figures, self.constant)

        value = self._round()
        figures = self.figures
        if value.adjusted() > self.value.adjusted():
            figures += 1

        return Quantity(value, figures, self.constant)

    def almosteq(self, other, rel_eps=None, abs_eps=None):
        """Determine if two ``Quantity`` objects are almost equal.

        Determine if two ``Quantity`` objects are approximately equal.
        Use ``mpmath.almosteq()`` to determine approximate equality.

        Unlike ``__eq__``, ``self`` and ``other`` are not rounded
        before checking equality, so ``self`` and ``other`` may differ
        after rounding.

        Parameters
        ----------
        other : sigcalc.Quantity
            A second ``Quantity`` to compare with ``self``.
        rel_eps :
            Passed to ``mpmath.almosteq()`` as ``mpmath.mpf(rel_eps)``
            if present.  See the
            `mpmmath documentation <https://mpmath.org/doc/current/general.html#mpmath.almosteq>`_
            for details.
        abs_eps :
            Passed to ``mpmath.almosteq()`` as ``mpmath.mpf(abs_eps)``
            if present.  See the
            `mpmmath documentation <https://mpmath.org/doc/current/general.html#mpmath.almosteq>`_
            for details.

        Returns
        -------
        bool
            ``True`` if almost equal, ``False`` otherwise.
        """  # noqa: E501
        with mpmath.workdps(getcontext().prec):
            if rel_eps:
                rel_eps = mpmath.mpmathify(rel_eps)
            if abs_eps:
                abs_eps = mpmath.mpmathify(abs_eps)

            if self.constant and other.constant:
                return all(
                    (
                        mpmath.almosteq(
                            mpmath.mpmathify(self.value),
                            mpmath.mpmathify(other.value),
                            rel_eps,
                            abs_eps,
                        ),
                    )
                )

            return all(
                (
                    mpmath.almosteq(
                        mpmath.mpmathify(self.value),
                        mpmath.mpmathify(other.value),
                        rel_eps,
                        abs_eps,
                    ),
                    self.figures == other.figures,
                )
            )

    # Exponential functions.

    def exp(self):
        """Calculate the base e exponential of a ``Quantity``.

        Uses the ``exp()`` function from ``decimal.Decimal`` for the
        exponential calculation and computes the significant figures
        from the significant figures of the input ``Quantity``.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> a = Quantity("1", "3")
        >>> b = Quantity("0", "3")
        >>> a.exp()
        Quantity("2.718281828459045235360287471", "2")
        >>> import warnings
        >>> warnings.simplefilter("error")
        >>> b.exp()  # warn on ambiguous input, but return an answer
        Traceback (most recent call last):
        ...
        RuntimeWarning: ambiguous zero valued ``Quantity``:  Quantity("0", "3")
        >>> warnings.simplefilter("ignore")
        >>> b.exp()
        Quantity("1", "0")
        >>> warnings.simplefilter("error")
        >>> c = Quantity("10", "2")
        >>> c.exp()  # warn on insufficient precision, but return an answer
        Traceback (most recent call last):
        ...
        RuntimeWarning: insufficient input precision in Quantity("10", "2"): the 2 abscissa digits consume the available precision (2)
        >>> warnings.simplefilter("ignore")
        >>> c.exp()
        Quantity("22026.46579480671651695790065", "0")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed exponential and
            significant figures and constant value from the input
            ``Quantity``.

        Raises
        ------
        RuntimeWarning
            If ``self.value`` is zero, there are no significant
            figures.  Raise and return ``Quantity("1", "0")``.  Also
            raise if there are fewer significant figures than abscissa
            digits since the abscissa digits are placeholders from the
            logarithm.
        """  # noqa: E501
        # Warn on zero.
        if self.value == Zero:
            warn(f"ambiguous zero valued ``Quantity``:  {repr(self)}", RuntimeWarning)
            return Quantity("1", "0")

        value = self.value.exp()

        # Default precision for no abscissa.
        figures = self.figures

        # Constants.
        if self.constant:
            return Quantity(
                value,
                constant=self.constant,
            )

        # Chop abscissa places since they are not significant in the logarithm.
        if abs(self.value) >= One:
            figures = self.figures - (self.value.adjusted() + One)

        if figures < One:
            warn(
                f"insufficient input precision in {repr(self)}: "
                f"the {self.value.adjusted() + One} "
                f"abscissa digits consume the available precision ({self.figures})",
                RuntimeWarning,
            )
            figures = Zero

        # No abscissa.
        return Quantity(
            value,
            figures,
        )

    def exp10(self):  # dead: disable
        """Calculate the base 10 exponential of a ``Quantity``.

        Uses the ``__pow__()`` function from ``decimal.Decimal`` for
        the exponential calculation and computes the significant
        figures from the significant figures of the input
        ``Quantity``.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> a = Quantity("1", "3")
        >>> b = Quantity("0", "3")
        >>> a.exp10()
        Quantity("10", "2")
        >>> import warnings
        >>> warnings.simplefilter("error")
        >>> b.exp10()  # warn on ambiguous input, but return an answer
        Traceback (most recent call last):
        ...
        RuntimeWarning: ambiguous zero valued ``Quantity``:  Quantity("0", "3")
        >>> warnings.simplefilter("ignore")
        >>> b.exp10()
        Quantity("1", "0")
        >>> warnings.simplefilter("error")
        >>> c = Quantity("10", "2")
        >>> c.exp10()  # warn on insufficient precision, but return an answer
        Traceback (most recent call last):
        ...
        RuntimeWarning: insufficient input precision in Quantity("10", "2"): the 2 abscissa digits consume the available precision (2)
        >>> warnings.simplefilter("ignore")
        >>> c.exp10()
        Quantity("10000000000", "0")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed exponential and
            significant figures and constant value from the input
            ``Quantity``.

        Raises
        ------
        RuntimeWarning
            If ``self.value`` is zero, there are no significant
            figures.  Raise and return ``Quantity("1", "0")``.  Also
            raise if there are fewer significant figures than abscissa
            digits since the abscissa digits are placeholders from the
            logarithm.
        """  # noqa: E501
        # Warn on zero.
        if self.value == Zero:
            warn(f"ambiguous zero valued ``Quantity``:  {repr(self)}", RuntimeWarning)
            return Quantity("1", "0")

        value = pow(Ten, self.value)

        # Default precision for no abscissa.
        figures = self.figures

        # Constants.
        if self.constant:
            return Quantity(
                value,
                constant=self.constant,
            )

        # Chop abscissa places since they are not significant in the logarithm.
        if abs(self.value) >= One:
            figures = self.figures - (self.value.adjusted() + One)

        if figures < One:
            warn(
                f"insufficient input precision in {repr(self)}: "
                f"the {self.value.adjusted() + One} "
                f"abscissa digits consume the available precision ({self.figures})",
                RuntimeWarning,
            )
            figures = Zero

        # No abscissa.
        return Quantity(
            value,
            figures,
        )

    def sqrt(self):
        """Calculate the square root of a ``Quantity``.

        Uses the ``sqrt()`` function from ``decimal.Decimal`` for the
        square root calculation and computes the significant figures
        from the significant figures of the input ``Quantity``.

        >>> from sigcalc import Quantity
        >>> a = Quantity("4", "3")
        >>> b = Quantity("9", "4")
        >>> a.sqrt()
        Quantity("2", "3")
        >>> b.sqrt()
        Quantity("3", "4")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed square root and
            significant figures and constant value from the input
            ``Quantity``.
        """
        return Quantity(self.value.sqrt(), self.figures, self.constant)

    # Logarithmic functions.

    def ln(self):
        """Calculate the natural logarithm of a ``Quantity``.

        Uses the ``ln()`` function from ``decimal.Decimal`` for the
        natural logarithm calculation and computes the significant
        figures from the significant figures of the input
        ``Quantity``.

        >>> from sigcalc import Quantity
        >>> a = Quantity("2", "3")
        >>> b = Quantity("3", "3")
        >>> a.ln()
        Quantity("0.6931471805599453094172321215", "3")
        >>> b.ln()
        Quantity("1.098612288668109691395245237", "4")

        Beware the usual problems with computing logarithms, as raised
        by the ``decimal`` module.  Precision is gained if the
        absolute value of the result is greater than or equal to one
        since the significant figures of the input determine the
        significant decimal places of the output.

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed natural logarithm,
            computed significant figures from the input and the size
            of the computed logarithm, and constant value from the
            input ``Quantity``.
        """
        value = self.value.ln()
        figures = self.figures

        if abs(value) >= One:
            # Significant figures includes the abscissa digits.
            figures += value.adjusted() + One

        if self.constant:
            return Quantity(value, constant=self.constant)

        return Quantity(value, figures)

    def log10(self):
        """Calculate the base 10 logarithm of a ``Quantity``.

        Uses the ``log10()`` function from ``decimal.Decimal`` for the
        base 10 logarithm calculation and computes the significant
        figures from the significant figures of the input
        ``Quantity``.

        >>> from sigcalc import Quantity
        >>> a = Quantity("100", "3")
        >>> a.log10()
        Quantity("2", "4")
        >>> b = Quantity("0.5", "3")
        >>> b.log10()
        Quantity("-0.3010299956639811952137388947", "3")

        Beware the usual problems with computing logarithms, as raised
        by the ``decimal`` module.  Precision is gained if the
        absolute value of the result is greater than or equal to one
        since the significant figures of the input determine the
        significant decimal places of the output.

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed base 10 logarithm,
            computed significant figures from the input and the size
            of the computed logarithm, and constant value from the
            input ``Quantity``.
        """
        value = self.value.log10()
        figures = self.figures

        if abs(value) >= One:
            # Significant figures includes the abscissa digits.
            figures += value.adjusted() + One

        if self.constant:
            return Quantity(value, constant=self.constant)

        return Quantity(value, figures)

    # Trigonometric functions and inverses.

    def sin(self):
        """Calculate the sine of a ``Quantity``.

        Uses ``mpmath.sin()`` to calculate the sine and computes the
        significant figures from the significant figures of the input
        ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> import mpmath
        >>> from sigcalc import Quantity
        >>> from sigcalc import pi
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> zero = Quantity("0", "3")
        >>> ninety = pi / Quantity("2", "3")
        >>> one_eighty = pi * Quantity("1", "3")
        >>> two_seventy = pi * Quantity("3", "3") / Quantity("2", "3")
        >>> three_sixty = pi * Quantity("2", "3")
        >>> zero.sin()
        Quantity("0.0", "3")
        >>> ninety.sin()
        Quantity("1.0", "3")
        >>> mpmath.almosteq(mpmath.mpmathify(one_eighty.sin().value), mpmath.mpmathify(Decimal("0")), 1e-25)
        True
        >>> two_seventy.sin()
        Quantity("-1.0", "3")
        >>> mpmath.almosteq(mpmath.mpmathify(three_sixty.sin().value), mpmath.mpmathify(Decimal("0")), 1e-25)
        True

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed sine and significant
            figures and constant value from the input ``Quantity``.
        """  # noqa: E501
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.sin(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def asin(self):
        """Calculate the inverse sine of a ``Quantity``.

        Uses ``mpmath.asin()`` to calculate the sine and computes the
        significant figures from the significant figures of the input
        ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> a = Quantity("0", "3")
        >>> b = Quantity("1", "3")
        >>> c = Quantity("-1", "3")
        >>> a.asin()
        Quantity("0.0", "3")
        >>> b.asin()
        Quantity("1.570796326794896619231321692", "3")
        >>> c.asin()
        Quantity("-1.570796326794896619231321692", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse sine and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.asin(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def cos(self):
        """Calculate the cosine of a ``Quantity``.

        Uses ``mpmath.cos()`` to calculate the cosine and computes the
        significant figures from the significant figures of the input
        ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from sigcalc import pi
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> zero = Quantity("0", "3")
        >>> ninety = pi / Quantity("2", "3")
        >>> one_eighty = pi * Quantity("1", "3")
        >>> two_seventy = pi * Quantity("3", "3") / Quantity("2", "3")
        >>> three_sixty = pi * Quantity("2", "3")
        >>> zero.cos()
        Quantity("1.0", "3")
        >>> mpmath.almosteq(mpmath.mpmathify(ninety.cos().value), mpmath.mpmathify(Decimal("0")), 1e-25)
        True
        >>> one_eighty.cos()
        Quantity("-1.0", "3")
        >>> mpmath.almosteq(mpmath.mpmathify(two_seventy.cos().value), mpmath.mpmathify(Decimal("0")), 1e-25)
        True
        >>> three_sixty.cos()
        Quantity("1.0", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed cosine and
            significant figures and constant value from the input
            ``Quantity``.
        """  # noqa: E501
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.cos(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def acos(self):
        """Calculate the inverse cosine of a ``Quantity``.

        Uses ``mpmath.acos()`` to calculate the cosine and computes the
        significant figures from the significant figures of the input
        ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> a = Quantity("0", "3")
        >>> b = Quantity("1", "3")
        >>> c = Quantity("-1", "3")
        >>> a.acos()
        Quantity("1.570796326794896619231321692", "3")
        >>> b.acos()
        Quantity("0.0", "3")
        >>> c.acos()
        Quantity("3.141592653589793238462643383", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse cosine and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.acos(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def tan(self):
        """Calculate the tangent of a ``Quantity``.

        Uses ``mpmath.tan()`` to calculate the tangent and computes the
        significant figures from the significant figures of the input
        ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from sigcalc import pi
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> zero = Quantity("0", "3")
        >>> forty_five = pi / Quantity("4", "3")
        >>> three_fifteen = pi * Quantity("7", "3") / Quantity("4", "3")
        >>> zero.tan()
        Quantity("0.0", "3")
        >>> forty_five.tan()
        Quantity("1.0", "3")
        >>> mpmath.almosteq(mpmath.mpmathify(three_fifteen.tan().value), mpmath.mpmathify(Decimal("-1")), 1e-25)
        True

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed tangent and
            significant figures and constant value from the input
            ``Quantity``.
        """  # noqa: E501
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.tan(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def atan(self):
        """Calculate the inverse tangent of a ``Quantity``.

        Uses ``mpmath.atan()`` to calculate the tangent and computes the
        significant figures from the significant figures of the input
        ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> a = Quantity("0", "3")
        >>> b = Quantity("1", "3")
        >>> c = Quantity("-1", "3")
        >>> a.atan()
        Quantity("0.0", "3")
        >>> b.atan()
        Quantity("0.7853981633974483096156608458", "3")
        >>> c.atan()
        Quantity("-0.7853981633974483096156608458", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse tangent and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.atan(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def csc(self):
        """Calculate the cosecant of a ``Quantity``.

        Uses ``mpmath.csc()`` to calculate the cosecant and computes
        the significant figures from the significant figures of the
        input ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> import mpmath
        >>> from sigcalc import Quantity
        >>> from sigcalc import pi
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> ninety = pi / Quantity("2", "3")
        >>> two_seventy = pi * Quantity("3", "3") / Quantity("2", "3")
        >>> ninety.csc()
        Quantity("1.0", "3")
        >>> two_seventy.csc()
        Quantity("-1.0", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed cosecant and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.csc(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def acsc(self):
        """Calculate the inverse cosecant of a ``Quantity``.

        Uses ``mpmath.acsc()`` to calculate the cosecant and computes
        the significant figures from the significant figures of the
        input ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> b = Quantity("1", "3")
        >>> c = Quantity("-1", "3")
        >>> b.acsc()
        Quantity("1.570796326794896619231321692", "3")
        >>> c.acsc()
        Quantity("-1.570796326794896619231321692", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse cosecant and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.acsc(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def sec(self):
        """Calculate the secant of a ``Quantity``.

        Uses ``mpmath.sec()`` to calculate the secant and computes the
        significant figures from the significant figures of the input
        ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from sigcalc import pi
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> zero = Quantity("0", "3")
        >>> one_eighty = pi * Quantity("1", "3")
        >>> three_sixty = pi * Quantity("2", "3")
        >>> zero.sec()
        Quantity("1.0", "3")
        >>> one_eighty.sec()
        Quantity("-1.0", "3")
        >>> three_sixty.sec()
        Quantity("1.0", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed secant and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.sec(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def asec(self):
        """Calculate the inverse secant of a ``Quantity``.

        Uses ``mpmath.asec()`` to calculate the secant and computes
        the significant figures from the significant figures of the
        input ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> b = Quantity("1", "3")
        >>> c = Quantity("-1", "3")
        >>> b.asec()
        Quantity("0.0", "3")
        >>> c.asec()
        Quantity("3.141592653589793238462643383", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse secant and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.asec(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def cot(self):
        """Calculate the cotangent of a ``Quantity``.

        Uses ``mpmath.cot()`` to calculate the cotangent and computes
        the significant figures from the significant figures of the
        input ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from sigcalc import pi
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> forty_five = pi / Quantity("4", "3")
        >>> ninety = pi / Quantity("2", "3")
        >>> one_thirty_five = pi * Quantity("3", "3") / Quantity("4", "3")
        >>> forty_five.cot()
        Quantity("1.0", "3")
        >>> mpmath.almosteq(mpmath.mpmathify(ninety.cot().value), mpmath.mpmathify(Decimal("0")), 1e-25)
        True
        >>> mpmath.almosteq(mpmath.mpmathify(one_thirty_five.cot().value), mpmath.mpmathify(Decimal("-1")), 1e-25)
        True

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed cotangent and
            significant figures and constant value from the input
            ``Quantity``.
        """  # noqa: E501
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.cot(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def acot(self):
        """Calculate the inverse cotangent of a ``Quantity``.

        Uses ``mpmath.acot()`` to calculate the cotangent and computes
        the significant figures from the significant figures of the
        input ``Quantity``.  This wrapper to manages the precision of
        ``mpmath`` to compute at the required precision of the current
        ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import Decimal
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> a = Quantity("0", "3")
        >>> b = Quantity("1", "3")
        >>> c = Quantity("-1", "3")
        >>> a.acot()
        Quantity("1.570796326794896619231321692", "3")
        >>> b.acot()
        Quantity("0.7853981633974483096156608458", "3")
        >>> c.acot()
        Quantity("-0.7853981633974483096156608458", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse cotangent and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.acot(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    # Hyperbolic functions and inverses.

    def sinh(self):
        """Calculate the hyperbolic sine of a ``Quantity``.

        Uses ``mpmath.sinh()`` to calculate the hyperbolic sine and
        computes the significant figures from the significant figures
        of the input ``Quantity``.  This wrapper to manages the
        precision of ``mpmath`` to compute at the required precision
        of the current ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> Quantity("0", "3").sinh()
        Quantity("0.0", "3")
        >>> Quantity("1", "3").sinh()
        Quantity("1.175201193643801456882381851", "3")
        >>> Quantity("-1", "3").sinh()
        Quantity("-1.175201193643801456882381851", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed hyperbolic sine and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.sinh(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def asinh(self):
        """Calculate the inverse hyperbolic sine of a ``Quantity``.

        Uses ``mpmath.asinh()`` to calculate the hyperbolic sine and
        computes the significant figures from the significant figures
        of the input ``Quantity``.  This wrapper to manages the
        precision of ``mpmath`` to compute at the required precision
        of the current ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> Quantity("0", "3").asinh()
        Quantity("0.0", "3")
        >>> Quantity("1", "3").asinh()
        Quantity("0.881373587019543025232609325", "3")
        >>> Quantity("-1", "3").asinh()
        Quantity("-0.881373587019543025232609325", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse hyperbolic
            sine and significant figures and constant value from the
            input ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.asinh(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def cosh(self):
        """Calculate the hyperbolic cosine of a ``Quantity``.

        Uses ``mpmath.cosh()`` to calculate the hyperbolic cosine and
        computes the significant figures from the significant figures
        of the input ``Quantity``.  This wrapper to manages the
        precision of ``mpmath`` to compute at the required precision
        of the current ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> Quantity("0", "3").cosh()
        Quantity("1.0", "3")
        >>> Quantity("1", "3").cosh()
        Quantity("1.543080634815243778477905621", "3")
        >>> Quantity("-1", "3").cosh()
        Quantity("1.543080634815243778477905621", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed hyperbolic cosine and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.cosh(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def acosh(self):
        """Calculate the inverse hyperbolic cosine of a ``Quantity``.

        Uses ``mpmath.acosh()`` to calculate the hyperbolic cosine and
        computes the significant figures from the significant figures
        of the input ``Quantity``.  This wrapper to manages the
        precision of ``mpmath`` to compute at the required precision
        of the current ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> Quantity("1", "3").acosh()
        Quantity("0.0", "3")
        >>> Quantity("2", "3").acosh()
        Quantity("1.316957896924816708625046347", "3")
        >>> Quantity("3", "3").acosh()
        Quantity("1.76274717403908605046521865", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse hyperbolic
            cosine and significant figures and constant value from the
            input ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.acosh(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def tanh(self):
        """Calculate the hyperbolic tangent of a ``Quantity``.

        Uses ``mpmath.tanh()`` to calculate the hyperbolic tangent and
        computes the significant figures from the significant figures
        of the input ``Quantity``.  This wrapper to manages the
        precision of ``mpmath`` to compute at the required precision
        of the current ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> Quantity("-1", "3").tanh()
        Quantity("-0.7615941559557648881194582826", "3")
        >>> Quantity("0", "3").tanh()
        Quantity("0.0", "3")
        >>> Quantity("1", "3").tanh()
        Quantity("0.7615941559557648881194582826", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed hyperbolic tangent
            and significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.tanh(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def atanh(self):
        """Calculate the inverse hyperbolic tangent of a ``Quantity``.

        Uses ``mpmath.atanh()`` to calculate the hyperbolic tangent
        and computes the significant figures from the significant
        figures of the input ``Quantity``.  This wrapper to manages
        the precision of ``mpmath`` to compute at the required
        precision of the current ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> Quantity("-0.5", "3").atanh()
        Quantity("-0.5493061443340548456976226185", "3")
        >>> Quantity("0", "3").atanh()
        Quantity("0.0", "3")
        >>> Quantity("0.5", "3").atanh()
        Quantity("0.5493061443340548456976226185", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse hyperbolic
            tangent and significant figures and constant value from
            the input ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.atanh(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def csch(self):
        """Calculate the hyperbolic cosecant of a ``Quantity``.

        Uses ``mpmath.csch()`` to calculate the hyperbolic cosecant
        and computes the significant figures from the significant
        figures of the input ``Quantity``.  This wrapper to manages
        the precision of ``mpmath`` to compute at the required
        precision of the current ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> Quantity("-1", "3").csch()
        Quantity("-0.8509181282393215451338427633", "3")
        >>> Quantity("1", "3").csch()
        Quantity("0.8509181282393215451338427633", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed hyperbolic cosecant
            and significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.csch(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def acsch(self):
        """Calculate the inverse hyperbolic cosecant of a ``Quantity``.

        Uses ``mpmath.acsch()`` to calculate the hyperbolic cosecant
        and computes the significant figures from the significant
        figures of the input ``Quantity``.  This wrapper to manages
        the precision of ``mpmath`` to compute at the required
        precision of the current ``decimal`` context.

        >>> from sigcalc import Quantity
        >>> from decimal import getcontext
        >>> getcontext().prec = 28
        >>> Quantity("-1", "3").acsch()
        Quantity("-0.881373587019543025232609325", "3")
        >>> Quantity("1", "3").acsch()
        Quantity("0.881373587019543025232609325", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse hyperbolic
            cosecant and significant figures and constant value from
            the input ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.acsch(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def sech(self):
        """Calculate the hyperbolic secant of a ``Quantity``.

        Uses ``mpmath.sech()`` to calculate the hyperbolic secant and
        computes the significant figures from the significant figures
        of the input ``Quantity``.  This wrapper to manages the
        precision of ``mpmath`` to compute at the required precision
        of the current ``decimal`` context.

        >>> from decimal import getcontext
        >>> from sigcalc import Quantity
        >>> getcontext().prec = 28
        >>> Quantity("-1", "3").sech()
        Quantity("0.6480542736638853995749773532", "3")
        >>> Quantity("0", "3").sech()
        Quantity("1.0", "3")
        >>> Quantity("1", "3").sech()
        Quantity("0.6480542736638853995749773532", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed hyperbolic secant and
            significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.sech(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def asech(self):
        """Calculate the inverse hyperbolic secant of a ``Quantity``.

        Uses ``mpmath.asech()`` to calculate the hyperbolic secant and
        computes the significant figures from the significant figures
        of the input ``Quantity``.  This wrapper to manages the
        precision of ``mpmath`` to compute at the required precision
        of the current ``decimal`` context.

        >>> from decimal import getcontext
        >>> from sigcalc import Quantity
        >>> getcontext().prec = 28
        >>> Quantity("0.5", "3").asech()
        Quantity("1.316957896924816708625046347", "3")
        >>> Quantity("1", "3").asech()
        Quantity("0.0", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse hyperbolic
            secant and significant figures and constant value from the
            input ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.asech(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def coth(self):
        """Calculate the hyperbolic cosecant of a ``Quantity``.

        Uses ``mpmath.coth()`` to calculate the hyperbolic cotangent
        and computes the significant figures from the significant
        figures of the input ``Quantity``.  This wrapper to manages
        the precision of ``mpmath`` to compute at the required
        precision of the current ``decimal`` context.

        >>> from decimal import getcontext
        >>> from sigcalc import Quantity
        >>> getcontext().prec = 28
        >>> Quantity("-1", "3").coth()
        Quantity("-1.313035285499331303636161247", "3")
        >>> Quantity("1", "3").coth()
        Quantity("1.313035285499331303636161247", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed hyperbolic cotangent
            and significant figures and constant value from the input
            ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.coth(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )

    def acoth(self):
        """Calculate the inverse hyperbolic cotangent of a ``Quantity``.

        Uses ``mpmath.acoth()`` to calculate the hyperbolic cotangent
        and computes the significant figures from the significant
        figures of the input ``Quantity``.  This wrapper to manages
        the precision of ``mpmath`` to compute at the required
        precision of the current ``decimal`` context.

        >>> from decimal import getcontext
        >>> from sigcalc import Quantity
        >>> getcontext().prec = 28
        >>> Quantity("-2", "3").acoth()
        Quantity("-0.5493061443340548456976226185", "3")
        >>> Quantity("2", "3").acoth()
        Quantity("0.5493061443340548456976226185", "3")

        Returns
        -------
        sigcalc.Quantity
            A new ``Quantity`` with the computed inverse hyperbolic
            cotangent and significant figures and constant value from
            the input ``Quantity``.
        """
        mpmath.mp.dps = getcontext().prec
        return Quantity(
            Decimal(
                mpmath.nstr(mpmath.acoth(mpmath.mpmathify(self.value)), mpmath.mp.dps)
            ),
            self.figures,
            self.constant,
        )



# ******************************************************************************
#
# sigcalc, significant figures calculations
#
# Copyright 2023-2024 Jeremy A Gray <gray@flyquackswim.com>.
#
# All rights reserved.
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# ******************************************************************************

"""Constant tests."""

from decimal import Decimal

import sigcalc


def test_properties_of_e():
    """Should have the correct properties."""
    assert sigcalc.e.value == Decimal("2.7182818284590452353602874713526624977572")
    assert sigcalc.e.figures == Decimal("41")
    assert sigcalc.e.constant is True


def test_properties_of_pi():
    """Should have the correct properties."""
    assert sigcalc.pi.value == Decimal("3.1415926535897932384626433832795028841971")
    assert sigcalc.pi.figures == Decimal("41")
    assert sigcalc.pi.constant is True



# ******************************************************************************
#
# sigcalc, significant figures calculations
#
# Copyright 2023-2024 Jeremy A Gray <gray@flyquackswim.com>.
#
# All rights reserved.
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# ******************************************************************************

"""Quantity hypothesis tests."""

import math
from decimal import ROUND_05UP
from decimal import ROUND_CEILING
from decimal import ROUND_DOWN
from decimal import ROUND_FLOOR
from decimal import ROUND_HALF_DOWN
from decimal import ROUND_HALF_EVEN
from decimal import ROUND_HALF_UP
from decimal import ROUND_UP
from decimal import Decimal
from decimal import InvalidOperation
from decimal import getcontext

import mpmath
import pytest
from hypothesis import HealthCheck
from hypothesis import assume
from hypothesis import given
from hypothesis import settings
from hypothesis import strategies as st
from hypothesis.strategies import composite

from sigcalc import Quantity
from sigcalc import pi

# Reusable decimal constants.
NegThousand = Decimal("-1000")
NegTen = Decimal("-10")
NegOne = Decimal("-1")
Zero = Decimal("0")
One = Decimal("1")
Two = Decimal("2")
Ten = Decimal("10")
Hundred = Decimal("100")
Thousand = Decimal("1000")


@composite
def quantities(draw, min_value=None, max_value=None):
    """Generate quantities."""
    return Quantity(
        draw(
            st.decimals(
                allow_nan=False,
                allow_infinity=False,
                min_value=min_value,
                max_value=max_value,
            )
        ),
        draw(
            st.integers(
                min_value=1,
                max_value=100,
            )
        ),
        draw(st.booleans()),
    )


@composite
def rounding(draw):
    """Generate a rounding mode."""
    return draw(
        st.sampled_from(
            [
                ROUND_CEILING,
                ROUND_DOWN,
                ROUND_FLOOR,
                ROUND_HALF_DOWN,
                ROUND_HALF_EVEN,
                ROUND_HALF_UP,
                ROUND_UP,
                ROUND_05UP,
            ]
        )
    )


@given(
    st.decimals(
        allow_nan=False,
        allow_infinity=False,
    ),
)
def test___init__raises_without_figures_nonconstant(value):
    """Should raise ``TypeError`` without ``figures`` if not constant."""
    with pytest.raises(TypeError):
        Quantity(value)


@given(
    st.decimals(
        allow_nan=False,
        allow_infinity=False,
    ),
    st.integers(
        min_value=1,
        max_value=100,
    ),
)
def test___init__creates_quantity_objects(value, figures):
    """Should create ``Quantity`` objects."""
    q = Quantity(value, Decimal(str(figures)))
    assert q.value == value
    assert q.figures == Decimal(str(figures))
    assert q.constant is False


@given(
    st.decimals(
        allow_nan=False,
        allow_infinity=False,
    ),
    st.integers(
        min_value=1,
        max_value=100,
    ),
)
def test___init__creates_quantity_constants(value, figures):
    """Should create ``Quantity`` objects."""
    q = Quantity(value, Decimal(str(figures)), constant=True)
    assert q.value == value
    assert q.figures == getcontext().prec
    assert q.constant is True


@given(quantities(), rounding())
def test_equality_hypothesis(q, r):
    """Should order ``Quantity`` objects."""
    getcontext().rounding = r
    assert q == q
    assert q <= q
    assert q >= q
    assert not q != q


@given(quantities(), rounding())
def test_ordering_hypothesis(q, r):
    """Should order ``Quantity`` objects."""
    # Zeroes satisfy equality.
    if r == ROUND_DOWN:
        assume(abs(q.value) >= 1)
    else:
        assume(q.value > 0)
    getcontext().rounding = r
    assert q + abs(q) > q
    assert q > q - abs(q)
    assert q < q + abs(q)
    assert q - abs(q) < q


@given(quantities(), rounding())
def test_equality_different_precision_hypothesis(q, r):
    """Should order ``Quantity`` objects."""
    getcontext().rounding = r
    if q.figures == getcontext().prec:
        p = Quantity(q.value, q.figures - One, False)
    else:
        p = Quantity(q.value, q.figures + One, q.constant)

    assert q != p


@given(quantities(), rounding())
def test_almost_equal_self_hypothesis(one, r):
    """Should be almost equal to self."""
    assert one.almosteq(one)


@given(
    quantities(
        min_value=Zero,
        max_value=Ten,
    ),
    rounding(),
)
def test_almost_equal_pass_eps_hypothesis(one, r):
    """Should pass both epsilons."""
    delta = mpmath.mpmathify("1e-25")
    abs_eps = "1"
    rel_eps = "1"

    two = Quantity(
        one.value + delta,
        one.figures,
        constant=one.constant,
    )
    assert one.almosteq(two, rel_eps=rel_eps)
    assert one.almosteq(two, abs_eps=abs_eps)
    assert one.almosteq(two, rel_eps=rel_eps, abs_eps=abs_eps)


@given(quantities(), quantities(), rounding())
def test_almost_equal_hypothesis(one, two, r):
    """Should determine approximate equality."""
    if one.figures == two.figures:
        eps = pow(mpmath.mpf("2"), -getcontext().prec + 4)
        if mpmath.mpmathify(abs(one.value - two.value)) > eps:
            assert not one.almosteq(two)
        else:
            assert one.almosteq(two)
    else:
        assert not one.almosteq(two)


@given(quantities(), rounding())
def test_abs_hypothesis(q, r):
    """Should calculate absolute value of ``Quantity`` objects."""
    getcontext().rounding = r
    p = Quantity(abs(q.value), q.figures, q.constant)
    assert abs(q) == p
    assert abs(abs(q)) == p


@given(quantities(), rounding())
def test_neg_hypothesis(q, r):
    """Should negate ``Quantity`` objects."""
    getcontext().rounding = r
    assert -(-q) == q


@given(quantities(), rounding())
def test_pos_hypothesis(q, r):
    """Should return ``Quantity`` objects."""
    getcontext().rounding = r
    p = Quantity(q.value, q.figures, q.constant)
    assert +q == q
    assert ++q == q
    assert +q == p


@given(quantities(), rounding())
def test_round_hypothesis(q, r):
    """Should round ``Quantity`` objects."""
    getcontext().rounding = r
    assert q.round() == q.round().round()


@given(quantities(), rounding())
def test_additive_identity_hypothesis(q, r):
    """``Quantity`` objects should have an additive identity."""
    getcontext().rounding = r
    id = Quantity("0", "1", True)
    assert q + id == q
    assert id + q == q
    assert id + q + id == q


@given(quantities(), rounding())
def test_subtractive_identity_hypothesis(q, r):
    """``Quantity`` objects should have a subtractive identity."""
    getcontext().rounding = r
    id = Quantity("0", "1", True)
    assert q - id == q


@given(quantities(), rounding())
def test_additive_inverse_hypothesis(q, r):
    """``Quantity`` objects should have an additive inverse."""
    getcontext().rounding = r
    # assert q - q == Quantity("0", q.figures, q.constant)
    assert q + q - q == q
    assert q - q + q == q


@given(quantities(), rounding())
def test_multiplicative_identity_hypothesis(q, r):
    """``Quantity`` objects should have a multiplicative identity."""
    getcontext().rounding = r
    id = Quantity("1", "1", True)
    assert q * id == q
    assert id * q == q
    assert id * q * id == q

    id = Quantity("1", q.figures)
    assert q * id == q
    assert id * q == q
    assert id * q * id == q


@given(quantities(), rounding())
def test_divisive_identity_hypothesis(q, r):
    """``Quantity`` objects should have a divisive identity."""
    getcontext().rounding = r
    id = Quantity("1", "1", True)
    assert q / id == q

    id = Quantity("1", q.figures)
    assert q / id == q


@given(quantities(), rounding(), st.booleans())
def test_constant_setter_hypothesis(q, r, c):
    """Should set the ``constant`` attribute."""
    getcontext().rounding = r
    q.constant = c
    assert q.constant is c


# Exponential function tests.
@given(quantities(), rounding())
def test_exp_hypothesis(q, r):
    """Should calculate the exponential of ``Quantity`` objects."""
    # Really need to round-trip with ln, since this repeats the tested
    # logic for significance.

    # Avoid decimal overflow.
    assume(q.value < Thousand)
    assume(q.value > NegThousand)

    # Avoid ambiguous zero warning.
    assume(q.value != Zero)

    # Ensure sufficient precision.
    assume(q.figures > (q.value.adjusted() + One))

    getcontext().rounding = r

    actual = q.exp()

    figures = q.figures

    if q.constant:
        expected = Quantity(
            q.value.exp(),
            constant=q.constant,
        )
    elif abs(q.value) >= One:
        # Magnitude of abscissa is greater than one; chop abscissa places.
        figures = q.figures - (q.value.adjusted() + One)

    expected = Quantity(
        q.value.exp(),
        figures,
        constant=q.constant,
    )

    assert actual.almosteq(expected)
    assert actual.constant == expected.constant


@given(
    st.integers(
        min_value=1,
        max_value=100,
    ),
    st.booleans(),
    rounding(),
)
def test_exp_ambiguous_zero_hypothesis(figures, constant, r):
    """Should warn on ambiguous zero input."""
    getcontext().rounding = r

    q = Quantity("0", figures, constant)
    expected = Quantity("1", "0")

    with pytest.warns(RuntimeWarning):
        actual = q.exp()

    assert actual == expected


@settings(
    suppress_health_check=[
        HealthCheck.filter_too_much,
        HealthCheck.too_slow,
    ],
)
@given(quantities(), rounding())
def test_exp_insufficient_precision_hypothesis(q, r):
    """Should warn on insufficient precision."""
    # Avoid decimal overflow.
    assume(q.value < Thousand)
    assume(q.value > NegThousand)

    # Avoid ambiguous zero warning.
    assume(q.value != Zero)

    # Ensure insufficient precision.
    assume(q.figures <= (q.value.adjusted() + One))

    getcontext().rounding = r

    with pytest.warns(RuntimeWarning):
        actual = q.exp()

    expected = Quantity(
        q.value.exp(),
        Zero,
    )

    assert actual.almosteq(expected)
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_exp10_hypothesis(q, r):
    """Should calculate the base 10 exponential of ``Quantity`` objects."""
    # Really need to round-trip with ln, since this repeats the tested
    # logic for significance.

    # Avoid decimal overflow.
    assume(q.value < Thousand)
    assume(q.value > NegThousand)

    # Avoid ambiguous zero warning.
    assume(q.value != Zero)

    # Ensure sufficient precision.
    assume(q.figures > (q.value.adjusted() + One))

    getcontext().rounding = r

    actual = q.exp10()

    figures = q.figures

    if q.constant:
        expected = Quantity(
            pow(Ten, q.value),
            constant=q.constant,
        )
    elif abs(q.value) >= One:
        # Magnitude of abscissa is greater than one; chop abscissa places.
        figures = q.figures - (q.value.adjusted() + One)

    expected = Quantity(
        pow(Ten, q.value),
        figures,
        constant=q.constant,
    )

    assert actual.almosteq(expected)
    assert actual.constant == expected.constant


@given(
    st.integers(
        min_value=1,
        max_value=100,
    ),
    st.booleans(),
    rounding(),
)
def test_exp10_ambiguous_zero_hypothesis(figures, constant, r):
    """Should warn on ambiguous zero input."""
    getcontext().rounding = r

    q = Quantity("0", figures, constant)
    expected = Quantity("1", "0")

    with pytest.warns(RuntimeWarning):
        actual = q.exp10()

    assert actual == expected


@settings(
    suppress_health_check=[
        HealthCheck.filter_too_much,
        HealthCheck.too_slow,
    ],
)
@given(quantities(), rounding())
def test_exp10_insufficient_precision_hypothesis(q, r):
    """Should warn on insufficient precision."""
    # Avoid decimal overflow.
    assume(q.value < Thousand)
    assume(q.value > NegThousand)

    # Avoid ambiguous zero warning.
    assume(q.value != Zero)

    # Ensure insufficient precision.
    assume(q.figures <= (q.value.adjusted() + One))

    getcontext().rounding = r

    with pytest.warns(RuntimeWarning):
        actual = q.exp10()

    expected = Quantity(
        pow(Ten, q.value),
        Zero,
    )

    assert actual.almosteq(expected)
    assert actual.constant == expected.constant


@given(quantities(), quantities(), rounding())
def test_power_hypothesis(base, exp, r):
    """Should calculate powers of ``Quantity`` objects."""
    # Avoid indeterminancy of zeroes.
    assume(base.value != Zero)

    # Avoid overflow/underflow.
    assume(abs(base.value) < Thousand)
    assume(abs(exp.value) < Hundred)

    getcontext().rounding = r

    if base.value < Zero and Decimal(int(exp.value)) != exp.value:
        with pytest.raises(InvalidOperation):
            actual = pow(base, exp)
    else:
        actual = pow(base, exp)

        if base.constant:
            expected = Quantity(
                pow(base.value, exp.value),
                constant=base.constant,
            )
        else:
            expected = Quantity(
                pow(base.value, exp.value),
                base.figures,
            )

        assert actual.almosteq(expected)
        assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_sqrt_hypothesis(q, r):
    """Should calculate the square root of ``Quantity`` objects."""
    # Avoid domain problems.
    assume(q.value > Zero)

    getcontext().rounding = r
    e = q.sqrt()

    assert q.value.sqrt() == e.value
    assert q.figures == e.figures
    assert q.constant == e.constant


# Logarithmic function tests.
@given(quantities(), rounding())
def test_ln_hypothesis(q, r):
    """Should calculate the natural logarithm of ``Quantity`` objects."""
    # Avoid logarithm domain problems.
    assume(q.value > Zero)
    assume(not math.isnan(q.value))

    getcontext().rounding = r
    actual = q.ln()

    # Calculate significant figures.
    if not q.constant and (abs(q.value) >= One.exp() or abs(q.value) <= NegOne.exp()):
        # Include abscissa digits.
        a = q.value.ln().adjusted() + 1
    else:
        # No abscissa digits.
        a = 0

    expected = Quantity(
        q.value.ln(),
        q.figures + a,
        q.constant,
    )

    assert actual.almosteq(expected)
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_log10_hypothesis(q, r):
    """Should calculate the base 10 logarithm of ``Quantity`` objects."""
    # Avoid logarithm domain problems.
    assume(q.value > Zero)
    assume(not math.isnan(q.value))

    getcontext().rounding = r

    actual = q.log10()

    # Calculate significant figures.
    if not q.constant and (abs(q.value) >= Ten or abs(q.value) <= Decimal("0.1")):
        # Include abscissa digits.
        a = q.value.log10().adjusted() + One
    else:
        # No abscissa digits.
        a = Zero

    expected = Quantity(
        q.value.log10(),
        q.figures + a,
        q.constant,
    )

    assert actual.almosteq(expected)
    assert actual.constant == expected.constant


# Logarithmic/exponential round trip tests.
@given(quantities(), rounding())
def test_ln_exp_hypothesis(expected, mode):
    """Should round trip exponential of natural logarithm."""
    # Avoid logarithm domain problems.
    assume(expected.value > Zero)
    # Avoid ambiguous zero.
    assume(expected.value != One)

    # Set rounding context.
    getcontext().rounding = mode

    actual = expected.ln().exp()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_exp_ln_hypothesis(expected, mode):
    """Should round trip natural logarithm of exponential."""
    # Avoid decimal overflow.
    assume(expected.value < Thousand)
    assume(expected.value > NegThousand)

    # Avoid ambiguous zero warning.
    assume(expected.value != Zero)

    # Ensure sufficient precision.
    assume(expected.figures > (expected.value.adjusted() + One))

    # Set rounding context.
    getcontext().rounding = mode

    actual = expected.exp().ln()

    assert actual.almosteq(expected, "1e-25")
    assert actual.constant == expected.constant


# Trigonometric function tests.
@given(quantities(), rounding())
def test_sin_hypothesis(q, r):
    """Should calculate the sine of ``Quantity`` objects."""
    e = q.sin()

    # Duplication; no need to test mpmath.
    assert (
        Decimal(mpmath.nstr(mpmath.sin(mpmath.mpmathify(q.value)), mpmath.mp.dps))
        == e.value
    )
    assert q.figures == e.figures
    assert q.constant == e.constant


@given(quantities(), rounding())
def test_asin_hypothesis(q, r):
    """Should calculate the inverse sine of ``Quantity`` objects."""
    assume(q.value >= -1 and q.value <= 1)
    e = q.asin()

    # Duplication; no need to test mpmath.
    assert (
        Decimal(mpmath.nstr(mpmath.asin(mpmath.mpmathify(q.value)), mpmath.mp.dps))
        == e.value
    )
    assert q.figures == e.figures
    assert q.constant == e.constant


@given(quantities(), rounding())
def test_asin_of_sin_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value >= -pi.value / Two and expected.value <= pi.value / Two)
    actual = expected.sin().asin()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_sin_of_asin_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value >= -1 and expected.value <= 1)
    actual = expected.asin().sin()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_cos_hypothesis(q, r):
    """Should calculate the cosine of ``Quantity`` objects."""
    e = q.cos()

    # Duplication; no need to test mpmath.
    assert (
        Decimal(mpmath.nstr(mpmath.cos(mpmath.mpmathify(q.value)), mpmath.mp.dps))
        == e.value
    )
    assert q.figures == e.figures
    assert q.constant == e.constant


@given(quantities(), rounding())
def test_acos_hypothesis(q, r):
    """Should calculate the inverse cosine of ``Quantity`` objects."""
    assume(q.value >= -1 and q.value <= 1)
    e = q.acos()

    # Duplication; no need to test mpmath.
    assert (
        Decimal(mpmath.nstr(mpmath.acos(mpmath.mpmathify(q.value)), mpmath.mp.dps))
        == e.value
    )
    assert q.figures == e.figures
    assert q.constant == e.constant


@given(quantities(), rounding())
def test_acos_of_cos_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value >= Zero and expected.value <= pi.value)
    actual = expected.cos().acos()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_cos_of_acos_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value >= -1 and expected.value <= 1)
    actual = expected.acos().cos()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_tan_hypothesis(q, r):
    """Should calculate the tangent of ``Quantity`` objects."""
    e = q.tan()

    # Duplication; no need to test mpmath.
    assert (
        Decimal(mpmath.nstr(mpmath.tan(mpmath.mpmathify(q.value)), mpmath.mp.dps))
        == e.value
    )
    assert q.figures == e.figures
    assert q.constant == e.constant


@given(quantities(), rounding())
def test_atan_hypothesis(q, r):
    """Should calculate the inverse tangent of ``Quantity`` objects."""
    e = q.atan()

    # Duplication; no need to test mpmath.
    assert (
        Decimal(mpmath.nstr(mpmath.atan(mpmath.mpmathify(q.value)), mpmath.mp.dps))
        == e.value
    )
    assert q.figures == e.figures
    assert q.constant == e.constant


@given(quantities(), rounding())
def test_atan_of_tan_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value >= -pi.value / Two and expected.value <= pi.value / Two)
    actual = expected.tan().atan()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_tan_of_atan_hypothesis(expected, r):
    """Should return input."""
    actual = expected.atan().tan()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_csc_hypothesis(q, r):
    """Should calculate the cosecant of ``Quantity`` objects."""
    assume(q.value >= -pi.value and q.value <= pi.value)
    assume(q.value != Zero)

    actual = q.csc()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.csc(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(quantities(), rounding())
def test_acsc_hypothesis(q, r):
    """Should calculate the inverse cosecant of ``Quantity`` objects."""
    assume(q.value >= One or q.value <= NegOne)

    actual = q.acsc()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.acsc(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(quantities(), rounding())
def test_acsc_of_csc_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value >= -pi.value / Two and expected.value <= pi.value / Two)
    assume(expected.value != Zero)

    actual = expected.csc().acsc()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_csc_of_acsc_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value <= -1 or expected.value >= 1)
    actual = expected.acsc().csc()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_sec_hypothesis(q, r):
    """Should calculate the secant of ``Quantity`` objects."""
    assume(q.value >= Zero and q.value <= pi.value)
    assume(q.value != pi.value / Two)

    actual = q.sec()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.sec(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(quantities(), rounding())
def test_asec_hypothesis(q, r):
    """Should calculate the inverse secant of ``Quantity`` objects."""
    assume(q.value >= One or q.value <= NegOne)

    actual = q.asec()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.asec(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(quantities(), rounding())
def test_asec_of_sec_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value >= Zero and expected.value <= pi.value)
    assume(expected.value != pi.value / Two)

    actual = expected.sec().asec()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_sec_of_asec_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value <= -1 or expected.value >= 1)
    actual = expected.asec().sec()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_cot_hypothesis(q, r):
    """Should calculate the cotangent of ``Quantity`` objects."""
    assume(q.value > Zero and q.value < pi.value)

    actual = q.cot()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.cot(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(quantities(), rounding())
def test_acot_hypothesis(q, r):
    """Should calculate the inverse cotangent of ``Quantity`` objects."""
    e = q.acot()

    # Duplication; no need to test mpmath.
    assert (
        Decimal(mpmath.nstr(mpmath.acot(mpmath.mpmathify(q.value)), mpmath.mp.dps))
        == e.value
    )
    assert q.figures == e.figures
    assert q.constant == e.constant


@given(
    quantities(
        min_value=Decimal("0.0000000001"),
        max_value=pi.value - Decimal("0.0000000001"),
    ),
    rounding(),
)
def test_acot_of_cot_hypothesis(expected, r):
    """Should return input."""
    assume(
        mpmath.acot(mpmath.cot(mpmath.mpmathify(expected.value))) > 0
        and mpmath.acot(mpmath.cot(mpmath.mpmathify(expected.value)))
        < mpmath.mpmathify(pi.value)
    )

    actual = expected.cot().acot()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_cot_of_acot_hypothesis(expected, r):
    """Should return input."""
    actual = expected.acot().cot()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


# Hyperbolic function tests.
@given(quantities(), rounding())
def test_sinh_hypothesis(q, r):
    """Should calculate the hyperbolic sine of ``Quantity`` objects."""
    # Avoid decimal overflow/underflow.
    assume(q.value > NegThousand and q.value < Thousand)

    actual = q.sinh()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.sinh(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(quantities(), rounding())
def test_asinh_hypothesis(q, r):
    """Should calculate the inverse hyperbolic sine of ``Quantity`` objects."""
    actual = q.asinh()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.asinh(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(quantities(), rounding())
def test_asinh_of_sinh_hypothesis(expected, r):
    """Should return input."""
    # Avoid decimal overflow/underflow.
    assume(expected.value > NegThousand and expected.value < Thousand)

    actual = expected.sinh().asinh()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(quantities(), rounding())
def test_sinh_of_asinh_hypothesis(expected, r):
    """Should return input."""
    actual = expected.asinh().sinh()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_cosh_hypothesis(q, r):
    """Should calculate the hyperbolic cosine of ``Quantity`` objects."""
    actual = q.cosh()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.cosh(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=One,
        max_value=Thousand,
    ),
    rounding(),
)
def test_acosh_hypothesis(q, r):
    """Should calculate the inverse hyperbolic cosine of ``Quantity`` objects."""
    actual = q.acosh()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.acosh(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=One,
        max_value=Thousand,
    ),
    rounding(),
)
def test_acosh_of_cosh_hypothesis(expected, r):
    """Should return input."""
    actual = expected.cosh().acosh()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(
    quantities(
        min_value=One,
        max_value=Thousand,
    ),
    rounding(),
)
def test_cosh_of_acosh_hypothesis(expected, r):
    """Should return input."""
    actual = expected.acosh().cosh()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_tanh_hypothesis(q, r):
    """Should calculate the hyperbolic tangent of ``Quantity`` objects."""
    actual = q.tanh()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.tanh(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=Decimal("-0.99"),
        max_value=Decimal("0.99"),
    ),
    rounding(),
)
def test_atanh_hypothesis(q, r):
    """Should calculate the inverse hyperbolic tangent of ``Quantity`` objects."""
    actual = q.atanh()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.atanh(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=NegTen,
        max_value=Ten,
    ),
    rounding(),
)
def test_atanh_of_tanh_hypothesis(expected, r):
    """Should return input."""
    actual = expected.tanh().atanh()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(
    quantities(
        min_value=NegOne,
        max_value=One,
    ),
    rounding(),
)
def test_tanh_of_atanh_hypothesis(expected, r):
    """Should return input."""
    actual = expected.atanh().tanh()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_csch_hypothesis(q, r):
    """Should calculate the hyperbolic cosecant of ``Quantity`` objects."""
    # Singularity.
    assume(q.value != Zero)

    actual = q.csch()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.csch(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_acsch_hypothesis(q, r):
    """Should calculate the inverse hyperbolic cosecant of ``Quantity`` objects."""
    # Singularity.
    assume(q.value != Zero)

    actual = q.acsch()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.acsch(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_acsch_of_csch_hypothesis(expected, r):
    """Should return input."""
    # Singularity.
    assume(expected.value != Zero)

    actual = expected.csch().acsch()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_csch_of_acsch_hypothesis(expected, r):
    """Should return input."""
    # Singularity.
    assume(expected.value != Zero)

    actual = expected.acsch().csch()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_sech_hypothesis(q, r):
    """Should calculate the hyperbolic secant of ``Quantity`` objects."""
    actual = q.sech()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.sech(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=Decimal("0.001"),
        max_value=Decimal("0.999"),
    ),
    rounding(),
)
def test_asech_hypothesis(q, r):
    """Should calculate the inverse hyperbolic secant of ``Quantity`` objects."""
    actual = q.asech()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.asech(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=Zero,
        max_value=Thousand,
    ),
    rounding(),
)
def test_asech_of_sech_hypothesis(expected, r):
    """Should return input."""
    actual = expected.sech().asech()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))


@given(
    quantities(
        min_value=Decimal("0.001"),
        max_value=Decimal("0.999"),
    ),
    rounding(),
)
def test_sech_of_asech_hypothesis(expected, r):
    """Should return input."""
    actual = expected.asech().sech()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_coth_hypothesis(q, r):
    """Should calculate the hyperbolic cotangent of ``Quantity`` objects."""
    # Singularity.
    assume(q.value != Zero)

    actual = q.coth()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.coth(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_acoth_hypothesis(q, r):
    """Should calculate the inverse hyperbolic cotangent of ``Quantity`` objects."""
    assume(q.value < NegOne or q.value > One)

    actual = q.acoth()
    expected = Quantity(
        Decimal(mpmath.nstr(mpmath.acoth(mpmath.mpmathify(q.value)), mpmath.mp.dps)),
        q.figures,
        constant=q.constant,
    )

    assert actual == expected


@given(
    quantities(
        min_value=NegTen,
        max_value=Ten,
    ),
    rounding(),
)
def test_acoth_of_coth_hypothesis(expected, r):
    """Should return input."""
    # Singularity.
    assume(expected.value != Zero)

    actual = expected.coth().acoth()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant


@given(
    quantities(
        min_value=NegThousand,
        max_value=Thousand,
    ),
    rounding(),
)
def test_coth_of_acoth_hypothesis(expected, r):
    """Should return input."""
    assume(expected.value < NegOne or expected.value > One)

    actual = expected.acoth().coth()

    assert actual.almosteq(expected, mpmath.mpf("1e-25"))
    assert actual.constant == expected.constant



# ******************************************************************************
#
# sigcalc, significant figures calculations
#
# Copyright 2023-2024 Jeremy A Gray <gray@flyquackswim.com>.
#
# All rights reserved.
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# ******************************************************************************

"""sigcalc constants."""

from decimal import Decimal

from .quantity import Quantity

pi = Quantity(
    Decimal("3.1415926535897932384626433832795028841971"),
    constant=True,
)
"""pi, as a constant, with forty decimal digits.

See `"How Many Decimals of Pi Do We Really Need?" <https://www.jpl.nasa.gov/edu/news/2016/3/16/how-many-decimals-of-pi-do-we-really-need/>`_ if you think you need more.
"""  # noqa: E501

e = Quantity(
    Decimal("2.7182818284590452353602874713526624977572"),
    constant=True,
)
"""Euler's number, as a constant, with forty decimal digits.

See `"How Many Decimals of Pi Do We Really Need?" <https://www.jpl.nasa.gov/edu/news/2016/3/16/how-many-decimals-of-pi-do-we-really-need/>`_ if you think you need more.
"""  # noqa: E501



# ******************************************************************************
#
# sigcalc, significant figures calculations
#
# Copyright 2023-2024 Jeremy A Gray <gray@flyquackswim.com>.
#
# All rights reserved.
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# ******************************************************************************

"""sigcalc module initialization."""

from .constants import e
from .constants import pi
from .quantity import Quantity
from .quantity import _most_significant_place
